#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ERMzilla
#
# Ein Programm zum Zeichnen von Entity Relationship Modellen.
#
# Detlev Ahlgrimm  14.11.2020 - 20.11.2020
#
# D.A.  21.11.2020  v1.02   Tests auf neueren wxPython-Versionen,
#                           Export vom ERMs als png zugefügt
# D.A.  22.11.2020  v1.03   Umzug der Border-Definition von RELATION nach RELPOS,
#                           Keyword TABCOL zugefügt
# D.A.  26.11.2020  v1.04   Tabellenabmessungen korrigiert bei langen Tabellennamen
# D.A.  28.11.2020  v1.05   leichte Umbauten, um die Tastatur zu aktivieren,
#                           sichtbarer Ausschnitt wird jetzt mit gedrückt gehaltener Strg-Taste bewegt,
#                           Markierung eines Ausschnitts führt zur Selektion der drunterliegenden Objekte,
#                           die Verschiebe-Aktion bezieht sich auf alle derzeit markierten Objekte
# D.A.  29.11.2020  v1.06   Aufgeräumt und Kommentare zugefügt
#

# getestet mit folgenden Versionen von wxPython:
#   wx.version() = "2.8.12.1 (gtk2-unicode)"                (unter "Python 2.7.12")
#   wx.version() = '3.0.2.0 gtk3 (classic)'                 (unter "Python 2.7.17")
#   wx.version() = '4.1.0 gtk3 (phoenix) wxWidgets 3.1.4'   (unter "Python 3.8.5") ... dazu in allen Dateien den Shebang auf python3 geändert!


import os
import sys
import wx

from Table       import Table
from Relation    import Relation
from FileOps     import FileOps
from globalStuff import *

VERSION="v1.06"

tables=[]           # hier landen die Table()-Objekte
relations=[]        # hier landen die Relation()-Objekte
filename=""         # Dateiname der aktuell geladenen Datei

# ----------------------------------------------------------------------
# Eine kleine Klasse, um neue Dateien via Drag&Drop öffnen zu können.
# In der aufrufenden Klasse muss die Funktion fileDropped() existieren.
class FileDrop(wx.FileDropTarget):
    def __init__(self, window):
        wx.FileDropTarget.__init__(self)
        self.window=window
        self.Ctrl_pressed=False

    def OnDropFiles(self, x, y, filenames):
        self.window.fileDropped(filenames)
        return True                                                     # will "Phoenix" so haben - sonst gibts Warnings

    def OnDragOver(self, x, y, d):
        if d==wx.DragCopy:
            if wx.GetKeyState(wx.WXK_CONTROL):
                self.Ctrl_pressed=True
                return(wx.DragLink)
            self.Ctrl_pressed=False
            return(wx.DragCopy)
        return(d)

# ----------------------------------------------------------------------
#
class SelectedObjekts():
    def __init__(self):
        self.clearAll()

    # ------------------------------------------------------------------
    # Löscht bzw. initialisiert die Menge der markierten Objekte.
    def clearAll(self):
        self.drag_start_mouse_pos=None
        self.obj={}                                                     # dict() mit set() als value
        self.base_pos={}                                                # dict() mit dict() als value

    # ------------------------------------------------------------------
    # Liefert True, wenn es markierte Objekte gibt.
    def hasElements(self):
        return len(self.obj)>0

    # ------------------------------------------------------------------
    # Überschreibt die Menge der markierten Objekte mit obj und data.
    # ...also bei sowas wie leftClick ohne Shift.
    def set(self, obj, data=None):
        self.obj={obj:{data}}

    # ------------------------------------------------------------------
    # Fügt das Objekt obj der Menge der markierten Objekte hinzu.
    def add(self, obj, data=None):
        if obj in self.obj:
            if data is not None:
                self.obj[obj].add(data)                                 # weiteren Eckpunkt zur bereits enthaltenen Relation zufügen
        else:
            self.obj.update({obj:{data}})                               # Eckpunkt oder Tabelle neu hinzufügen

    # ------------------------------------------------------------------
    # Fügt alle Eckpunkte gemäß data aus der Relation obj zur Menge der
    # markierten Objekte hinzu.
    # data enthält ein dict() mit den Index-Werten der Eckpunkte.
    def add_many(self, obj, data):
        self.obj.update({obj:data})

    # ------------------------------------------------------------------
    # Liefert True, wenn das Objekt obj zur Menge der markierten
    # Objekte gehört.
    # Wenn obj eine Tabelle ist, darf data nicht übergeben werden.
    # Wenn obj ein Eckpunkt ist, ist dessen Index in data zu übergeben.
    def find(self, obj, data=None):
        if obj in self.obj:
            if data is None:
                return True                                             # Tabelle gefunden
            elif data in self.obj[obj]:
                return True                                             # Eckpunkt gefunden
        return False

    # ------------------------------------------------------------------
    # Entfernt obj aus der Liste der markierten Objekte.
    # Wenn obj eine Tabelle ist, darf data nicht übergeben werden.
    # Wenn obj ein Eckpunkt ist, ist dessen Index in data zu übergeben.
    def remove(self, obj, data=None):
        if data is None:
            if obj in self.obj:
                self.obj.pop(obj)                                       # Tabelle aus der Menge der markierten Objekte löschen
        elif obj in self.obj:
            self.obj[obj].discard(data)                                 # Eckpunkt aus dem value-Set löschen
            if len(self.obj[obj])==0:                                   # wenn es dadurch leer geworden ist...
                self.obj.pop(obj)                                       # ...wird das gesamte Relations-Objekt aus der Menge der markierten Objekte entfernt

    # ------------------------------------------------------------------
    # Setzt die Start-Koordinate für die Move-Aktion und vermerkt die
    # derzeitigen Positionen aller vom Move betroffenen Objekte - um
    # während des Moves mit relativen Abständen zur ursprünglichen
    # Objekt-Position arbeiten zu können.
    def initMoveable(self, pos):
        self.drag_start_mouse_pos=pos
        if pos is None:
            self.base_pos={}
        else:
            for t in tables:
                if t in self.obj:
                    self.base_pos.update({t:t.pos})                     # ursprüngliche Tabellen-Positionen merken
            for r in relations:
                if r in self.obj:
                    self.base_pos[r]=dict()
                    for i in self.obj[r]:
                        self.base_pos[r].update({i:r.line[i]})          # ursprüngliche Eckpunkt-Positionen merken

    # ------------------------------------------------------------------
    # Liefert True, wenn es markierte Objekte gibt und eine
    # Start-Koordinate für die Move-Aktion gesetzt ist.
    def isReadyToBeMoved(self):
        return self.hasElements() and self.drag_start_mouse_pos is not None


# ----------------------------------------------------------------------
#
class ERMzillaWindow(wx.Window):
    def __init__(self, parent):
        wx.Window.__init__(self, parent)
        self.parent=parent
        self.InitUI()

    # ------------------------------------------------------------------
    # Fenter initialisieren
    def InitUI(self):
        self.ctrlIsDown=False                                           # True, wenn die Strg-Taste gedrückt ist
        self.shiftIsDown=False                                          # True, wenn die Shift-Taste gedrückt ist
        self.leftDownPos=None                                           # die Koordinate, an der die linke Maus-Taste geklickt wurde (nur während leftDown !=None)
        self.errors_in_file=None                                        # True, wenn die geladene Datei Fehler enthielt (zur Sperre von "save")
        self.lastPosForShift=None                                       # zur Move-Delta-Bestimmung während Ausschnitts-Verschiebungen
        self.selectionRect=None                                         # Auswahl-Rechteck (nur während leftDown !=None)
        self.selObj=SelectedObjekts()                                   # Datenstruktur zur Aufnahme der markierten Objekte

        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)

        self.statusbar=self.parent.CreateStatusBar(5)
        self.statusbar.SetStatusWidths([0, 100, 100, -1, -1])

        self.__createMenu()
        
        self.filename_dt=FileDrop(self)                                 # Drag&Drop von ermz-Dateien aktivieren
        self.SetDropTarget(self.filename_dt)

        self.Bind(wx.EVT_CONTEXT_MENU,  self.onContextMenu)
        self.Bind(wx.EVT_PAINT,         self.onPaint)
        self.Bind(wx.EVT_MOTION,        self.onMotion)
        self.Bind(wx.EVT_LEFT_UP,       self.onLeftUp)
        self.Bind(wx.EVT_LEFT_DOWN,     self.onLeftDown)
        self.Bind(wx.EVT_LEFT_DCLICK,   self.onLeftDclick)
        self.Bind(wx.EVT_MOUSEWHEEL,    self.onMouseWheel)
        self.Bind(wx.EVT_KEY_DOWN,      self.onKeyDown)
        self.Bind(wx.EVT_KEY_UP,        self.onKeyUp)

        self.Centre()
        wx.CallLater(10, self.SetFocus)
        wx.CallLater(50, self.loadFile)

    # ------------------------------------------------------------------
    # Stellt ein Raster auf dem dc dar.
    def showRaster(self, dc):
        dc.SetPen(wx.Pen("#eeeeee", width=1, style=wx.SOLID))
        sx, sy=dc.GetSize()
        for x in range(int(sx/RASTER)+1):
            dc.DrawLine(x*RASTER, 0, x*RASTER, sy)
        for y in range(int(sy/RASTER)+1):
            dc.DrawLine(0, y*RASTER, sx, y*RASTER)
        dc.SetPen(wx.BLACK_PEN)

    # ------------------------------------------------------------------
    # Lädt die Datei gemäß der globalen Variable filename.
    def loadFile(self):
        if filename=="" or not os.path.isfile(filename):
            return
        file_ops=FileOps(tables, relations)
        self.errors_in_file=not file_ops.load(filename)
        if len(file_ops.log)>0:
            lns=""
            for ln in file_ops.log:
                lns+=ln+"\n"
            wx.MessageBox(lns, "Warnings / Errors")
            file_ops.log=[]
        self.statusbar.SetStatusText("tables: %d"%(len(tables),), 1)
        self.statusbar.SetStatusText("relations: %d"%(len(relations),), 2)
        if len(tables)==0 and len(relations)==0:
            self.parent.SetTitle("ERMzilla %s  (empty or bad file)"%(VERSION,))
        else:
            self.parent.SetTitle("ERMzilla %s  (%s)"%(VERSION, filename))
        self.Refresh()

    # ------------------------------------------------------------------
    # Löscht die internen Datenstrukturen für Tabellen und Relationen.
    def unloadFile(self):
        global relations, tables
        for r in relations:
            del r
        for t in tables:
            del t
        relations=[]
        tables=[]

    # ------------------------------------------------------------------
    # Wird aus FileDrop aufgerufen, wenn ein oder mehrere DateiObjekte
    # auf dem Fenster "fallengelassen" wurden.
    def fileDropped(self, filenames):
        global filename
        if len(filenames)!=1:
            wx.MessageBox("dropping multiple files is not allowed", "Error")
            return
        self.unloadFile()
        filename=filenames[0]
        self.loadFile()

    # ------------------------------------------------------------------
    # Darstellung der Tabellen und Relationen.
    def onPaint(self, event):
        self.dc=wx.AutoBufferedPaintDC(self)
        self.__drawDC(self.dc)

    # ------------------------------------------------------------------
    # Darstellung der Tabellen und Relationen auf dc.
    # Wird auch von saveSnapshot() mit einem MemoryDC aufgerufen.
    def __drawDC(self, dc):
        dc.SetBackground(wx.Brush("#FFFFFF"))
        dc.Clear()
        dc.SetTextForeground("#000000")
        if self.raster_mode:
            self.showRaster(dc)                                         # ggf. Raster darstellen

        for t in tables:                                                # Tabellen darstellen
            t.calcSize(dc)
            t.drawTable(dc, t in self.selObj.obj)
        for r in relations:                                             # Relationen darstellen
            fl=r.getFullList()
            dc.DrawLines(fl)
            if r.border_start=="L":
                sx, _=dc.GetTextExtent(r.rel_type[0])
                dc.DrawText(r.rel_type[0], fl[0][0]-5-sx, fl[0][1]-10)
            else:
                dc.DrawText(r.rel_type[0], fl[0][0]+5, fl[0][1]-10)
            if r.border_end=="L":
                sx, _=dc.GetTextExtent(r.rel_type[1])
                dc.DrawText(r.rel_type[1], fl[-1][0]-5-sx, fl[-1][1]-10)
            else:
                dc.DrawText(r.rel_type[1], fl[-1][0]+5, fl[-1][1]-10)

            if self.rel_points_mode:                                    # ggf. Eckpunkte darstellen
                for p in fl:
                    dc.DrawCircle(p[0], p[1], 2)
            if r in self.selObj.obj:                                    # markierte Eckpunkte werden immer dargestellt
                data=self.selObj.obj[r]
                dc.SetPen(wx.Pen("#ff0000", 3))
                for i, p in enumerate(fl[1:]):
                    if i in data:
                        dc.DrawCircle(p[0], p[1], 2)
                dc.SetPen(wx.Pen("#000000", 1))
        if self.selectionRect:                                          # ggf. das Markierungs-Rechteck darstellen
            (x1, y1), (x2, y2)=self.selectionRect
            pen=dc.GetPen()
            dc.SetPen(wx.Pen("#0000ff"))
            dc.DrawLines([(x1,y1), (x2,y1), (x2,y2), (x1,y2), (x1,y1)])
            dc.SetPen(pen)

    # ------------------------------------------------------------------
    # Liefert True, wenn pos auf einem markierten Objekt/Element liegt.
    def __isClickPosOnMarkedObject(self, pos):
        for r in relations:
            if r in self.selObj.obj:                                    # markierte Relation gefunden
                i=r.isPointOnCornerMark(pos)
                if i in self.selObj.obj[r]:
                    return True                                         # pos liegt über einem markierten Eckpunkt dieser Relation
        for t in tables:
            if t in self.selObj.obj:
                if t.isUnderXY(pos):
                    return True                                         # pos liegt über einer markierten Tabelle
        return False

    # ------------------------------------------------------------------
    # Linke Maustaste wurde geklickt.
    def onLeftDown(self, event):
        self.leftDownPos=event.GetPosition()
        if self.selObj.hasElements() and not self.shiftIsDown:          # es gibt markierte Elemente, die Shift-Taste ist nicht gedrückt...
            if self.__isClickPosOnMarkedObject(self.leftDownPos):       # ...und der Klick war auf einem markierten Element
                self.selObj.initMoveable(self.leftDownPos)              # also sollen jetzt Objekte verschoben werden.
                return

        # testen, ob ein Objekt angeklickt/markiert wurde
        sel_done=False
        for r in relations:                                             # wurde ein Eckpunkt angeklickt?
            cm=r.isPointOnCornerMark(self.leftDownPos)
            if cm is not None:                                          # Eckpunkt wurde angeklickt
                sel_done=True
                if self.shiftIsDown:                                    # Shift-Taste ist gedrückt
                    if self.selObj.find(r, cm):                         # wenn der Eckpunkt bereits markiert war...
                        self.selObj.remove(r, cm)                       # ...Markierung weg
                        break
                    else:
                        self.selObj.add(r, cm)                          # ...sonst Markierung setzen
                        break
                else:
                    self.selObj.set(r, cm)                              # Eckpunkt wurde ohne Shift geklickt -> Einzel-Markierung setzen
                    break
        if not sel_done:                                                # wenn bei den Relationen nix gefunden wurde
            for t in tables:                                            # wurde eine Tabelle angeklickt?
                if t.isUnderXY(self.leftDownPos):                       # Tabelle wurde angeklickt
                    sel_done=True
                    if self.shiftIsDown:                                # Shift-Taste ist gedrückt
                        if self.selObj.find(t):                         # wenn die Tabelle bereits markiert war...
                            self.selObj.remove(t)                       # ...Markierung weg
                            break
                        else:
                            self.selObj.add(t)                          # ...sonst Markierung setzen
                            break
                    else:
                        self.selObj.set(t)                              # Tabelle wurde ohne Shift geklickt -> Einzel-Markierung setzen
                        break
        self.selObj.initMoveable(self.leftDownPos)
        if not sel_done and not self.shiftIsDown:                       # weder Eckpunkt noch Tabelle wurde (ohne Shift) angeklickt
            self.selObj.clearAll()                                      # also gibts jetzt keine markierten Objekte mehr

        self.lastPosForShift=self.leftDownPos                           # Verschiebung des Bildschim-Inhaltes vorbereiten

    # ------------------------------------------------------------------
    # Linke Maustaste wurde losgelassen.
    def onLeftUp(self, event):
        self.leftDownPos=None
        if self.selectionRect:                                          # wenn zuvor ein Bereich ausgewählt wurde...
            nr=normalizeRectangle(self.selectionRect)
            for r in relations:                                         # ...markierte Objekte ermitteln
                ip=r.getCornerMarksUnderRectangle(nr)
                if ip:
                    self.selObj.add_many(r, ip)
            for t in tables:
                if t.hasIntersectionWithRectangle(nr):
                    self.selObj.add(t)
        self.selectionRect=None
        for t in tables:                                                # Tabellen am Raster ausrichten
            t.pos=pointOnRaster(t.pos)
        for r in relations:                                             # und die Relationen ebenso
            for i, p in enumerate(r.line):
                r.line[i]=pointOnRaster(p)
        self.Refresh()

    # ------------------------------------------------------------------
    # Der MausCursor wird über den DC bewegt.
    def onMotion(self, event):
        cur_pos=event.GetPosition()
        self.statusbar.SetStatusText("pos: %s"%(cur_pos,), 4)
        if self.leftDownPos:
            if self.selObj.isReadyToBeMoved():                          # wenn markierte Objekte existieren -> diese bewegen
                x, y=cur_pos
                for t in tables:
                    if t in self.selObj.obj:
                        x1=x-(self.selObj.drag_start_mouse_pos[0]-self.selObj.base_pos[t][0])
                        y1=y-(self.selObj.drag_start_mouse_pos[1]-self.selObj.base_pos[t][1])
                        t.moveAbs((x1, y1))
                for r in relations:
                    if r in self.selObj.obj:
                        for i in self.selObj.obj[r]:
                            x1=x-(self.selObj.drag_start_mouse_pos[0]-self.selObj.base_pos[r][i][0])
                            y1=y-(self.selObj.drag_start_mouse_pos[1]-self.selObj.base_pos[r][i][1])
                            r.moveAbsCornerMarkOnIndex((x1, y1), i)
            elif self.ctrlIsDown:                                       # ...bei gedrückter Strg-Taste soll der sichtbare Ausschnitt bewegt werden
                scroll_distXY=cur_pos-self.lastPosForShift
                self.lastPosForShift=cur_pos
                self.__moveERM(scroll_distXY)
            else:                                                       # ...ansonsten soll ein Bereich markiert werden
                self.selectionRect=(self.leftDownPos, cur_pos)
            self.Refresh()

    # ------------------------------------------------------------------
    # Verschiebt das gesamte ERM um delta Pixel.
    def __moveERM(self, delta):
        for t in tables:
            t.moveRel(delta)
        for r in relations:
            for i, p in enumerate(r.line):
                r.line[i]=(p[0]+delta[0], p[1]+delta[1])

    # ------------------------------------------------------------------
    # Linke Maustaste wurde doppel-geklickt.
    # Doppelklick auf Relations-Linie fügt einen neuen Eckpunkt ein
    # Doppelklick auf Relations-Eckpunkt löscht den Eckpunkt
    def onLeftDclick(self, event):
        x, y=event.GetPosition()
        for r in relations:
            a=r.isPointOnThisRelation((x, y))
            if a:
                if a[0]=="LN":
                    r.insertCornerMarkAfterIndex((x, y), a[1])
                    self.Refresh()
                    break
                elif a[0]=="CM":
                    r.removeCornerMarkOnIndex(a[1])
                    self.selObj.clearAll()
                    self.Refresh()
                    break

    # ------------------------------------------------------------------
    # Wenn sich unter dem MausCursor ein TableConnector befindet, führt
    # das Drehen am Scrollrad zum Wechsel der Tabellen-Seite.
    def onMouseWheel(self, event):
        point=event.GetPosition()
        for r in relations:
            tc=r.isPointOnTableConnector(point)
            if tc:
                if tc[0]==0:    r.swapStartBorder()
                else:           r.swapEndBorder()
                self.Refresh()
                return

    # ------------------------------------------------------------------
    # Eine Taste wurde betätigt.
    def onKeyDown(self, event):
        if event.GetKeyCode()==wx.WXK_CONTROL:
            self.ctrlIsDown=True
        if event.GetKeyCode()==wx.WXK_SHIFT:
            self.shiftIsDown=True

    # ------------------------------------------------------------------
    # Eine Taste wurde losgelassen.
    def onKeyUp(self, event):
        if event.GetKeyCode()==wx.WXK_CONTROL:
            self.ctrlIsDown=False
        if event.GetKeyCode()==wx.WXK_SHIFT:
            self.shiftIsDown=False

    # ------------------------------------------------------------------
    # Stellt das PopUp-Menü dar.
    def onContextMenu(self, event):
        if os.path.isfile(filename) and not self.errors_in_file:
            self.menue.Enable(100, True)
        else:
            self.menue.Enable(100, False)
        self.PopupMenu(self.menue)
        self.Refresh()

    # ------------------------------------------------------------------
    # Erstellt das PopUp-Menü.
    def __createMenu(self):
        self.raster_mode=True
        self.rel_points_mode=True
        self.menue=wx.Menu()
        self.menue.Append(80, 'export snapshot')
        self.menue.Append(90, 'reload')
        self.menue.Append(100, 'save')
        self.menue.Append(500, 'raster on/off', "", True)
        self.menue.Append(510, 'relation points on/off', "", True)
        self.Bind(wx.EVT_MENU, self.men_export_snapshot, id=80)
        self.Bind(wx.EVT_MENU, self.men_reload, id=90)
        self.Bind(wx.EVT_MENU, self.men_save, id=100)
        self.Bind(wx.EVT_MENU, self.men_raster, id=500)
        self.menue.Check(500, self.raster_mode)
        self.Bind(wx.EVT_MENU, self.men_rel_points, id=510)
        self.menue.Check(510, self.rel_points_mode)
        if not os.path.isfile(filename):
            self.menue.Enable(100, False)

    # ------------------------------------------------------------------
    # Menü - export snapshot
    # Exportiert das ERM als PNG-Datei. Die Bild-Datei landet im selben
    # Verzeichnis wie die ERM-Datei - die Datei-Endung wird durch
    # png ersetzt.
    def men_export_snapshot(self, event):
        se=os.path.splitext(os.path.abspath(filename))
        fn=se[0]+".png"
        self.saveSnapshot(fn)

    # ------------------------------------------------------------------
    # Menü - reload
    def men_reload(self, event):
        self.unloadFile()
        self.loadFile()
        self.Refresh()

    # ------------------------------------------------------------------
    # Menü - save
    def men_save(self, event):
        strg_d=""
        strg_p=""
        for t in tables:
            sd, sp=t.write()
            strg_d+=sd
            strg_p+=sp
            for r in relations:
                if r.start_col.split(".")[0]==t.name:
                    sd, sp=r.write()
                    strg_d+=sd
                    strg_p+=sp
        with open(filename, "wt") as fl:
            if strg_d.startswith("\n"):
                fl.write(strg_d[1:])
            else:
                fl.write(strg_d)
            fl.write("\n\n\n")
            fl.write(strg_p)

    # ------------------------------------------------------------------
    # Menü - Raster an/aus
    def men_raster(self, event):
        self.raster_mode=not self.raster_mode

    # ------------------------------------------------------------------
    # Menü - Darstellung der Eckpunkte an/aus
    def men_rel_points(self, event):
        self.rel_points_mode=not self.rel_points_mode

    # ------------------------------------------------------------------
    # Liefert die äußeren Abmessungen des ERMs
    def getBoundariesERM(self):
        min_x=min_y=max_x=max_y=None
        for t in tables:
            if min_x:   min_x=min(min_x, t.pos[0])
            else:       min_x=t.pos[0]
            if min_y:   min_y=min(min_y, t.pos[1])
            else:       min_y=t.pos[1]
            if max_x:   max_x=max(max_x, t.pos[0]+t.size[0])
            else:       max_x=t.pos[0]+t.size[0]
            if max_y:   max_y=max(max_y, t.pos[1]+t.size[1])
            else:       max_y=t.pos[1]+t.size[1]
        # der Test auf None kann hier entfallen, weil es keine Relation
        # ohne Table gibt
        for r in relations:
            fl=r.getFullList()
            for p in fl:
                min_x=min(min_x, p[0])
                min_y=min(min_y, p[1])
                max_x=max(max_x, p[0])
                max_y=max(max_y, p[1])
        return (min_x, min_y, max_x, max_y)

    # ------------------------------------------------------------------
    # Speichert das ERM als <filename>.png.
    def saveSnapshot(self, filename):
        x1, y1, x2, y2=self.getBoundariesERM()                          # Abmessungen des ERMs bestimmen
        w=x2-x1+2*RASTER
        h=y2-y1+2*RASTER

        bmp=wx.EmptyBitmap(w, h)
        #bmp=wx.Bitmap(w, h)                                            # "Phoenix" mag EmptyBitmap() nicht mehr
        memDC=wx.MemoryDC()
        memDC.SelectObject(bmp)

        dx=dy=0
        if x1<0:        dx=abs(x1-RASTER)                               # es liegen Objekte links ausserhalb des aktuellen Bildausschnitts
        elif x1>RASTER: dx=-x1+RASTER                                   # links ist ungenutzte Fläche im aktuellen Bildausschnitt
        if y1<0:        dy=abs(y1-RASTER)                               # es liegen Objekte oberhalb des aktuellen Bildausschnitts
        elif y1>RASTER: dy=-y1+RASTER                                   # oben ist ungenutzte Fläche im aktuellen Bildausschnitt

        if dx!=0 or dy!=0:
            self.__moveERM((dx, dy))                                    # ERM ausrichten
        self.__drawDC(memDC)                                            # ERM darstellen
        if dx!=0 or dy!=0:
            self.__moveERM((-dx, -dy))                                  # zu vorheriger Einstellung zurückkehren

        img=bmp.ConvertToImage()
        img.SaveFile(filename, wx.BITMAP_TYPE_PNG)                      # Bild-Datei speichern


# ----------------------------------------------------------------------
# 
class ERMzillaFrame(wx.Frame):
    def __init__(self, parent, pos=wx.DefaultPosition, size=wx.DefaultSize):
        style=wx.DEFAULT_FRAME_STYLE
        wx.Frame.__init__(self, None, wx.ID_ANY, pos=pos, size=size, style=style)
        self.panel=ERMzillaWindow(self)

# ----------------------------------------------------------------------
# 
if __name__=='__main__':
    if len(sys.argv)>1:
        ff=os.path.isfile(sys.argv[1])
        if ff:
            filename=sys.argv[1]

    app=wx.App(False)
    frame=ERMzillaFrame(None, pos=(100, 100), size=(1350, 750))
    frame.Show(True)
    app.MainLoop()
